#ifndef DAQCHANNELMAP_H
#define DAQCHANNELMAP_H

#include "DAQChannelMap/DAQChannelMapBase.h"


#define DAQCHANNELMAP_VERSION 2


namespace daqchannelmap {      //!< DAQ Channel Map Package

#ifndef EVEN
#define EVEN(x) (0 == ((x) % 2)) ? 1 : 0
#endif

#ifndef ODD
#define ODD(x)  (1 == ((x) % 2)) ? 1 : 0
#endif


  //! Structure for the fully qualified Universal hardware identifier
  struct UniversalDAQChannel{
    int diblock_ID;		//!< Di-block identifier field
    int dcm_id;		        //!< DCM identifier field (di-block local)
    int dcmport_id;		//!< DCM Data port identifier
    int pixel_id;		//!< FEB/APD pixel identifier
  };

  struct UniversalDCMChannel{
    int diblock_ID;		//!< Di-block identifier field
    int dcm_ID;		        //!< DCM identifier field (di-block local)
    int detector_ID;		//!< Detector identifier (IPND,NDOS,Near,Far....)
  };

  /*! 
    The DAQ Channel Map Class is a singleton class 
    designed to provide translation functions between
    the DAQ numbering schemes and the physics (physical)
    layout of the detector channels.
  */

  class DAQChannelMap: public DAQChannelMapBase {

    public:

        static DAQChannelMap* getInstance(); //!< Method to obtain the class instance
        ~DAQChannelMap();		//!< Default Destructor

        /* Encoding Routines */

        dchan encodeDChan(lchan logicalchan);
        lchan encodeLChan(dchan daqchan);

        //! Encoding Routine to correctly pack the DAQ specific channel parameters into a single 32bit identifier
        dchan encodeDChan(DetectorIdent,
                          diblock_t,
                          dcm_id_t,
                          dcm_link_t,
                          pixel_t) const;

        //! Encoding Routine to correctly pack the Offline specific channel parameters into a single 32bit identifier
        lchan encodeLChan(DetectorIdent,
                          offline_plane_t,
                          offline_cell_t) const;

        /* Decoding Routines */

        //! Convience Methods for accessing DAQ parameters from the logical space
        dchan l2d(offline_plane_t, offline_cell_t, DetectorIdent);

        // Get stuff given daq channel
        DetectorIdent getDetector (dchan daqchan) const; //!< Decodes detector identity from universal channel number
        diblock_t     getDiBlock  (dchan daqchan) const; //!< Decodes Diblock number from universal daq channel number
        dcm_id_t      getDCM      (dchan daqchan) const; //!< Decodes DCM position on the diblock from universal daq channel number
        dcm_link_t    getDCMLink  (dchan daqchan) const; //!< Decodes Link Port on the DCM from the universal daq channel number
        pixel_t       getPixel    (dchan daqchan) const; //!< Decodes APD pixel number from the universal daq channel number
        DetView_TYPE  getView     (dchan daqchan) const; //!< Returns the View (X or Y) from the universal daq channel number

        offline_cell_t  getCell(dchan daqchan) const;                        //!< Returns the cell position, as referenced relative to the full detector, of the daq channel
        offline_plane_t getPlane(dchan daqchan) ;                            //!< Returns the plane position, as referenced relative to the full detector, of the daq channel
        uint32_t        getLocalModuleInDCM(dcm_link_t, DiBlock_TYPE) const; //!< Returns the module number (0 or 1) to which a given module belongs

        // Other lookup functions
        DiBlock_TYPE  getDiBlockType       (dchan daqchan) const;      //!< Returns the DiBlock type (i.e. AB, BA, AA) of the daq channel  -- Verified
        DiBlock_TYPE  getDiBlockTypeOffline(lchan logicalchan) const;  //!< Returns the DiBlock type (i.e. AB, BA, AA) of the offline channel  -- Verified
        DetBlock_TYPE getBlockType(lchan logicalchan) const;	       //!< Returns the DiBlock type (i.e. AB, BA, AA) of the daq channel  -- Verified
        DetBlock_TYPE getBlockType(block_t, DetectorIdent) const;      //!< Returns the DiBlock type (i.e. AB, BA, AA) of the daq channel  -- Verified

        // Offline translation routines (i.e. "logical" channels)
        // Basic
        DetectorIdent   getDetectorOffline(lchan logicalchan) const;   //!< Returns the detector that the offline channel belongs to                           -- Verified
        offline_plane_t getPlaneOffline(lchan logicalchan) const;      //!< Returns the plane (detector wide) that the offline channel belongs to              -- Verified
        offline_cell_t  getCellOffline(lchan logicalchan) const;       //!< Returns the cell the cell on the current plane that the offline channel belongs to -- Verified

        // Block/Diblock
        block_t         getBlockOffline(lchan logicalchan) const;              //!< Returns the Block that the offline channel is on                                   -- Verified
        diblock_t       getDiBlockOffline(lchan logicalchan) const;	       //!< Returns the DiBlock that the offline channel is on                                 -- Verified
        uint32_t        getLocalBlockInDiblockOffline(lchan logicalchan) const;//!< If it's the upstream block in the diblock, then it's 0, if not, it's 1

        // View
        DetView_TYPE    getViewOffline     (lchan logicalchan) const;     //!< Returns the Plane type that the offline channel is on
        DetPlane_TYPE   getPlaneTypeOffline(lchan logicalchan) const;     //!< Returns the View that the offline channel is on
        DetPlane_TYPE   getPlaneType(DetectorIdent, dcm_id_t, DiBlock_TYPE) const;

        // Plane Related
        offline_plane_t getLocalPlaneInDiblockOffline    (lchan logicalchan) const;//!< Returns the Plane relative to the start of the current diblock    -- Verified
        offline_plane_t getLocalPlaneInBlockOffline      (lchan logicalchan) const;  //!< Returns the Plane relative to the start of the current block      -- Verified
        offline_plane_t getLocalPlaneInDiblockViewOffline(lchan logicalchan);   //!< Returns the Plane number relative to the current view and diblock -- Verified
        offline_plane_t getLocalPlaneInDiblockView       (dcm_link_t, DiBlock_TYPE) const;   //!< Returns the Plane number relative to the current view and diblock -- Verified
        offline_plane_t getPlaneViewOffline              (lchan logicalchan);	         //!< Returns the Plane number relative to the current view             -- Verified

        
        dcm_id_t   getDCMOffline(lchan logicalchan) const;
        uint32_t   getLocalModuleInDCMOffline(lchan) const;   //!< Returns the module number (0 or 1) to which a given module belongs
        uint32_t   getLocalModuleInDCMOffline(online_cell_t, DiBlock_TYPE) const;


        uint32_t   getCellDCMLocalOffline(lchan logicalchan) const;
    
        pixel_t    getPixelOffline(lchan logicalchan) const;
        dcm_link_t getDCMLinkOffline(lchan logicalchan);
                                                
        /// Print Information about DAQ Channel
        bool        print       (dchan daqchan);
        /// Print Information about the Offline logical Channel
        bool        printOffline(lchan logicalchan);

    protected:

        DAQChannelMap(); //<! Private Constructor

    private:
        /// First DCM in the view. if the channel is in a YZ plane, the DCM is 1.
        /// Otherwise, it's 7 for the far detector and 2 for NDOS
        dcm_id_t        getFirstDCMInViewOffline(lchan logicalchan) const;
        dcm_id_t        getFirstDCMInView(DetectorIdent detector, DetPlane_TYPE, DiBlock_TYPE) const;

        dcm_id_t        getLastDCMInViewOffline(lchan logicalchan) const;
        dcm_id_t        getLastDCMInView(DetectorIdent detector, DetPlane_TYPE, DiBlock_TYPE) const;

        uint32_t        getLocalBlockInDiblock(DiBlock_TYPE, offline_plane_t local_plane_in_diblock_in_view, DetPlane_TYPE) ;

        offline_plane_t getLocalPlaneInDiblock(offline_plane_t local_plane_in_diblock_view, DiBlock_TYPE, DetPlane_TYPE) ;

        online_cell_t   getOnlineCellOffline(lchan logicalchan) const;
        online_cell_t   getOnlineCell(DetectorIdent, dcm_id_t, dcm_link_t, pixel_t, DiBlock_TYPE) const;

        online_cell_t   getLocalCellInDCM(online_cell_t, DiBlock_TYPE) const;


        /// Offline to Online and vice versa.
        offline_cell_t  onlineCellToOfflineCell(online_cell_t , DetectorIdent, DetPlane_TYPE) const;
        online_cell_t   offlineCellToOnlineCell(offline_cell_t, DetectorIdent, DetPlane_TYPE) const;
        /// If input cell if offline_cell, then the output is online cell
        uint32_t        ConvertCell(uint32_t input_cell, DetectorIdent, DetPlane_TYPE) const;

        bool            printBinary(uint32_t);
        bool            printHex(uint32_t);
        
    private:
         static DAQChannelMap*  _instanceID; //!< Instance pointer for singleton class


  };

} // endnamespace daqchannelmap

#endif /* DAQCHANNELMAP_H */
